%{
/**
*** Copyright (C) 1994 - 2002 Hewlett-Packard Company.
***
*** The information and source code contained herein is the exclusive
*** property of Hewlett-Packard Company and may not be disclosed, examined
*** or reproduced in whole or in part without explicit written authorization
*** from the company.
**/

/**
*** Author:  Geoffrey Brown, Paolo Faraboschi
**/

static char sccs_id[] = "$Id: grammar.y,v 1.6 2003/10/09 09:49:11 frb Exp $";

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>

#ifdef YYBISON
# ifdef __hpux
#  include <alloca.h>
# endif
#endif

#include "vextypes.h"

#include <iostream>
#include <vector>

#include <boost/config.hpp>
#include <boost/tuple/tuple.hpp>
#include <boost/graph/adjacency_list.hpp>
#include <boost/graph/graph_utility.hpp>
#include <boost/graph/graphviz.hpp>
 
using namespace std;
using namespace boost;
 
#include "sched.h"


extern int yylex (void);

enum { LOCAL = 0, GLOBAL = 1 };
static char *vexscope[] = {"local","global"};

void yyerror (char *s);
void yywarn (char *s);
int yywrap ();

static void print_exp(vexexp *exp);
static vexexp *build_binexp(char op, vexexp *s1, vexexp *s2);
static vexexp *build_unexp(char op, vexexp *s1);
static vexexp *build_strexp(char *s);
static vexexp *build_numexp(int x);

void print_oplist(oplist *list);
 
static vexexp *add_op(char *r, vexexp *expr);
static oplist *make_op(char *r, unsigned flags);
static oplist *concat_oplist(oplist *l1, oplist *l2); 
static oplist *set_def_flags(oplist *list);
static oplist *set_use_flags(oplist *list);
static vexinstr *make_vexinstr(char *as_op, VEX_flags f, oplist *ops);
static vexinstr *concat_vexinstr(vexinstr *l1, vexinstr *l2);
 
 
static int fcycle = 0;
static unsigned next_instr_id = 0;
 
/* the one and only scheduler object */
static sched the_sched;
 
%}

%start asm_file

%union
{
   int       val;
   char      *text;
   vexopc    *opc;
   vexexp    *exp;
   oplist    *ops;
   vexinstr  *instrs;
   vexfun    fun;
}

%token            _ALIGN
%token            _ASCII
%token            _BSS
%token <opc>      _CALL_JMP
%token            _COMMENT
%token            _COMMON
%token            _DATA
%token            _DATA1
%token            _DATA2
%token            _DATA4
%token            _ENDP
%token            _ENTRY
%token            _EQU
%token            _IMPORT
%token            _PROC
%token            _REAL4
%token            _REAL8
%token            _RETURN
%token            _RTA
%token            _SECTION
%token            _SKIP
%token            _SVERSION
%token            _TEXT
%token            _TRACE
%token            _TYPE

%token <text>     CLUST
%token <text>     NAME
%token <val>      NUMBER
%token <opc>      OPCODE
%token <opc>      XNOP
%token <text>     QUOTE_STRING
%token <text>     REGNAME

%type   <text>    call_jmp_tgt
%type   <text>    name
%type   <text>    arg
%type   <text>    ret

%type   <val>     data_size
%type   <val>     scope
%type   <val>     .dup

%type   <exp>     data_val
%type   <exp>     expr
%type   <ops>     mop_arg
%type   <ops>     mop_arglist
%type   <ops>     .mop_arglist

%type   <instrs>  mop
%type   <instrs>  mop_list
%type   <instrs>  .mop_list
%type   <instrs>  normal_mop
%type   <instrs>  bundle
%type   <instrs>  code_item
%type   <instrs>  code_list


%type   <text>    label_decl

%type   <fun>     entry_dir

%left '+' '-' 
%left USIGN

%left INIT
%left _SKIP

%left ARGS
%left CLUST

/* dummy tokens, used outside yacc */
%token            BLOCK
%token            BUNDLE
%token            EQU
%token            LABEL

%%

/********************************************************************************
 **     File structure
 ********************************************************************************/

asm_file        :       asm_section
                |       asm_file asm_section 
                ;

asm_section     :       _SECTION data_section
                |       _SECTION bss_section
                |       _SECTION text_section
                |       _COMMENT QUOTE_STRING { }
                |       _SVERSION name
                |       _RTA NUMBER
                ;

        /*------------------------------------------------------*/
        /*              data section                            */
        /*------------------------------------------------------*/

data_section    :       _DATA .data_dir_list 
                ;


.data_dir_list  :       /* empty */
                |       data_dir_list
                ;

data_dir_list   :       data_dir
                |       data_dir_list data_dir
                ;


data_dir        :       _ALIGN NUMBER                   { }
                |       _SKIP NUMBER                    { }
                |       name scope                      { }
                        init_list %prec INIT            { }
                |       equ_dir                         { }
                |       common_dir                      { }
                |       data_global_dir                 { }
                |       type_dir                        { }
                ;

init_list       :       init_dir
                |       init_list init_dir
                ;

init_dir        :       data_size data_val .dup         { }
                |       _SKIP NUMBER                    { }
                |       _ASCII QUOTE_STRING             { }
                ;

data_val        :       expr               { $$ = $1; }
                |       NUMBER ',' NUMBER  { $$ = build_numexp($1); }

data_size       :       _DATA1          { $$ = 1; }
                |       _DATA2          { $$ = 2; }
                |       _DATA4          { $$ = 4; }
                |       _REAL4          { $$ = 4; }
                |       _REAL8          { $$ = 8; }
                ;

.dup            :       /* empty */     { $$ = 0;  }
                |       ':' NUMBER      { $$ = $2; }
                ;


data_global_dir :       _IMPORT name    { }
                ;

        /*------------------------------------------------------*/
        /*              bss section                             */
        /*------------------------------------------------------*/

bss_section     :       _BSS .bss_dir_list
                ;

.bss_dir_list   :       /* empty */
                |       bss_dir_list
                ;

bss_dir_list    :       bss_dir
                |       bss_dir_list bss_dir
                ;

bss_dir         :       bss_label_decl
                |       common_dir
                |       equ_dir
                |       _ALIGN NUMBER { }
                |       _SKIP NUMBER  { }
                ;

bss_label_decl  :       name scope    { }
                ;

common_dir      :       _COMMON name ',' NUMBER ',' NUMBER 
                ;

        /*------------------------------------------------------*/
        /*              text section                            */
        /*------------------------------------------------------*/

text_section    :       _TEXT .text_inst_list
                ;

.text_inst_list :       /* empty */
                |       text_inst_list
                ;

text_inst_list  :       text_inst                   
                |       text_inst_list text_inst
                ;

text_inst       :       proc_inst     
                |       equ_dir
                |       text_global_dir
                |       type_dir
                |       label_decl
                |       _ALIGN NUMBER { }
;

proc_inst       :       proc entry_dir code_list _ENDP {
                          printf("scheduling function \"%s\"\n", $2.str);
                          vexinstr *istream = $3;

                          sched::ddg DDG;
                          the_sched.computeDDG($2.str, DDG, istream);
                          
                          char name[512];
                          snprintf(name, 512, "%s.dot", $2.str);
                          the_sched.dumpDDG(name, DDG);
                          std::list<vexinstr *> L;
                          the_sched.schedule(DDG, L);

                          printf("result:\n");
                          std::list<vexinstr *>::iterator I, E;
                          for(I=L.begin(), E = L.end(); I != E; ++I) {
                            vexinstr *p = *I;
                            if(p)
                              printf("\t%d: %s\n", p->id, p->as_op);
                            else
                              printf("\tNOP\n");
                          }
                          
                          next_instr_id = 0;
                        }
                |       proc name scope _ENDP {
                            printf("empty function \"%s (%s)\"\n", $2, vexscope[$3]);
                        }
                |       proc _ENDP
                ;

proc            :       _PROC { fcycle = 0; }
                ;

code_list       :       code_item              { $$ = $1; }
                |       code_list code_item    { $$ = concat_vexinstr($1, $2); }
                ;

code_item       :       bundle           { $$ = $1;   }
                |       call_jmp_dir     { $$ = NULL; }
                |       equ_dir          { $$ = NULL; }
                |       label_decl       {
                           $$ = make_vexinstr($1, (VEX_flags)0, NULL);
                           $$->is_label = true;
                        }
                |       return_dir       { $$ = NULL; }
                |       trace_dir        { $$ = NULL; }
                ;

trace_dir       :       _TRACE NUMBER
                |       _TRACE NUMBER ',' NUMBER 
                ;

label_decl      :       name scope  { $$ = $1; }
                ;

call_jmp_dir    :       _CALL_JMP call_jmp_tgt ',' callc ',' { }
                        arg '(' .arg_list ')' ','            { }
                        ret '(' .arg_list ')'                { }
                ;

call_jmp_tgt    :       name          { $$ = $1; }
                |       REGNAME       { $$ = $1; }
                |       /* empty */   { $$ = "(empty)"; }
                ;

entry_dir       :       _ENTRY callc ',' regloclist ',' arg '(' .arg_list ')' name scope { 
                            $$.str = $10;
                            $$.scope = $11;
                        }
                |       name scope  {
                            $$.str = $1;
                            $$.scope = $2;
                        }
                ;

callc           :       NAME  { }
                ;

return_dir      :       _RETURN  ret '(' .arg_list ')'  { }
                ;

arg             :       NAME  { $$ = $1; }
                ;

ret             :       NAME  { $$ = $1; }
                ;

text_global_dir :       _IMPORT name  { }
                ;

equ_dir         :       _EQU name ',' expr { }
                ;

scope           :       ':'          { $$ = LOCAL; }
                |       ':'':'       { $$ = GLOBAL; }
                ;

type_dir        :       _TYPE name ',' '@' NAME
                ;

/**********************************************************************************
 **     Pseudo-instruction argument list (call, return, entry)
 **********************************************************************************/

.arg_list       :       /* empty */
                |       arg_list 
                |       stackexp arg_list 
                ;

stackexp        :       expr ',' NUMBER ';'  { }
                ;

arg_list        :       arg_desc              
                |       arg_list ',' arg_desc 
                ;

arg_desc        :       expr ':' arg_type               { }
                |       expr ':' expr ':' arg_type      { }
                |       ':' expr ':' arg_type           { }
                ;

arg_type        :       name            { }
                |       '[' NUMBER ']'  { }
                ;

regloclist      :       regloc         
                |       regloclist ',' regloc 
                ;

regloc          :       NAME '=' expr  { }
                ;

/**********************************************************************************
 **     BUNDLE structure and MOP opcodes
 **********************************************************************************/

bundle          :      .mop_list  end_bundle  {
/*                            printf("%5d]  -------------------------\n", fcycle++); */
                           $$ = $1;
                        }
                ;

end_bundle      :       ';'';' 
                ;

.mop_list       :       /* empty */   { $$ = NULL; }
                |       mop_list      { $$ = $1; }
                ;

mop_list        :       mop           { $$ = $1; }
                |       mop_list mop  { $$ = concat_vexinstr($1, $2); }
                ;

mop             :       normal_mop    { $$ = $1;   }
		|	xnop_mop      { $$ = NULL; }
                |       asm_mop       { $$ = NULL; }
                ;

normal_mop      :       CLUST OPCODE .mop_arglist {
                          set_use_flags($3);
/*                           printf("        %-5s %s", $1, $2->as_op); */
/*                           print_oplist($3); */
/*                           printf("\n"); */
                          $$ = make_vexinstr($2->as_op, $2->flags, $3);
                        }
                ;

xnop_mop	:	XNOP NUMBER {
/*                           printf("        xnop %d\n", $2); */
                        }
	 	;

asm_mop         :       CLUST OPCODE ',' NUMBER .mop_arglist {
                          assert(0 && "not yet implemented");
                        }
                ;

/**********************************************************************************
 **     MOP argument list
 **********************************************************************************/

.mop_arglist    :       /* empty */   %prec ARGS    { $$ = NULL; }
                |       mop_arglist '=' mop_arglist { $$ = concat_oplist(set_def_flags($1), $3); }
                |       mop_arglist '=' %prec ARGS  { $$ = $1; }
                |       mop_arglist                 { $$ = $1; }
                ;



mop_arglist     :       mop_arg                     { $$ = $1; }
                |       mop_arglist ',' mop_arg     { $$ = concat_oplist($1, $3); }
                ;

mop_arg         :       expr { 
                            $$ = $1->ops;
                        }
                |       '[' REGNAME ']' {
                            $$ = make_op($2, OP_USE);
                        }
                |       expr '[' REGNAME ']' {
                            $$ = concat_oplist($1->ops, make_op($3, OP_USE));
                        }
                ;

/**********************************************************************************
 **     General add/sub arithmetic expression
 **********************************************************************************/

expr            :       '(' expr ')'          { $$ = $2; }
                |       expr '+' expr         { $$ = build_binexp('+', $1, $3); }
                |       expr '-' expr         { $$ = build_binexp('-', $1, $3); }
                |       '-' expr %prec USIGN  { $$ = build_unexp('-', $2); }
                |       '+' expr %prec USIGN  { $$ = build_unexp('+', $2); }
                |       '~' expr %prec USIGN  { $$ = build_unexp('~', $2); }
                |       name                  { $$ = build_strexp($1); }
                |       REGNAME               { $$ = add_op($1, build_strexp($1));}
                |       NUMBER                { $$ = build_numexp($1); }
                ;

/**********************************************************************************
 **     Generic name (could be NAME or OPCODE)
 **********************************************************************************/
name            :       NAME                  { $$ = $1; }
                |       OPCODE                { $$ = $1->as_op; }
                |       CLUST                 { $$ = $1; }
                ;
/**********************************************************************************/
/**********************************************************************************/

%%


void print_oplist(oplist *list) {  
  oplist *l = list;
  printf(" {");  
  while(l) {
    if(l != list) printf(", ");
    printf("%s[", l->reg);
    if(l->flags & OP_USE)
      printf("U");
    if(l->flags & OP_DEF)
      printf("D");
    printf("]");
    l = l->next;
  }
  printf("}");  
}

static void print_exp(vexexp *exp)
{
    int x = exp->val;
    char *s = exp->str;

    if (x && s)
        printf("(");
    if (s) {
        printf("%s", s);
        if (x > 0)
            printf(" + %d", x);
        if (x < 0)
            printf(" - %d", -x);
    }
    else
        printf("%d", x);
    if (x && s)
        printf(")");
}

static vexinstr *make_vexinstr(char *as_op, VEX_flags f, oplist *ops) {
  vexinstr *instr = (vexinstr*) malloc(sizeof(vexinstr));
  assert(instr && "out of memory");
  instr->id = next_instr_id ++;
  instr->as_op = strdup(as_op);
  instr->flags = f;
  instr->ops = ops;
  instr->is_label=false;
  instr->next = NULL;
  return instr;
}

static oplist *make_op(char *r, unsigned flags) {
  oplist *opl = (oplist*) malloc(sizeof(oplist));
  assert(opl && "out of memory");
  opl->next = NULL;
  opl->reg = strdup(r);
  opl->flags = flags;  
  return opl;
}

static vexexp *add_op(char *r, vexexp *expr) {
  oplist *opl = make_op(r, OP_UNKNOWN);
  opl->next = expr->ops;
  expr->ops = opl;
  return expr;  
}

static oplist *concat_oplist(oplist *l1, oplist *l2) {
  if(l1 && l2) {
    oplist *p = l1;
    while(p->next) p=p->next;
    p->next = l2;
    return l1;
  }
  else if(l1)
    return l1;
  else
    return l2;
}

static vexinstr *concat_vexinstr(vexinstr *l1, vexinstr *l2) {
  if(l1 && l2) {
    vexinstr *p = l1;
    while(p->next) p=p->next;
    p->next = l2;
    return l1;
  }
  else if(l1)
    return l1;
  else
    return l2;
}

static oplist *set_def_flags(oplist *list) {
  oplist *l = list;
  while(l) {
    if(l->flags == OP_UNKNOWN)
      l->flags |= OP_DEF;
    l = l->next;
  }
  return list;
}

static oplist *set_use_flags(oplist *list) {
  oplist *l = list;
  while(l) {
    if(l->flags == OP_UNKNOWN)
      l->flags |= OP_USE;
    l = l->next;
  }
  return list;
}

static vexexp *build_binexp(char op, vexexp *s1, vexexp *s2)
{
    int nval;
    int v1 = s1->val;
    int v2 = s2->val;
    vexexp *exp = (vexexp*) malloc(sizeof(vexexp));
    assert(exp && "out of memory");
    switch (op) {
        case '-': nval = v1 - v2; break;
        case '+': nval = v1 + v2; break;
        default:
            fprintf(stderr,"illegal binary op (`%c') at line %d\n", 
                op, yylineno);
            exit(3);
    }
    if (s1->str && s2->str) {
        fprintf(stderr,"illegal binary exp (%s, %s) at line %d\n", 
            s1->str, s2->str, yylineno);
    }
    exp->str = s1->str ? s1->str : s2->str;
    exp->val = nval;
    exp->ops = concat_oplist(s1->ops, s2->ops);
    return exp;
}

static vexexp *build_unexp(char op, vexexp *s1)
{
    int nval;
    vexexp *exp = (vexexp*) malloc(sizeof(vexexp));
    assert(exp && "out of memory");
    int v1 = s1->val;
    switch (op) {
        case '-': nval = -v1; break;
        case '+': nval = +v1; break;
        case '~': nval = ~v1; break;
        default:
            fprintf(stderr,"illegal unary op (`%c') at line %d\n", 
                op, yylineno);
            exit(3);
    }
    if (s1->str) {
        fprintf(stderr,"illegal unary exp (%s) at line %d\n", 
            s1->str, yylineno);
    }
    exp->val = nval;
    exp->ops = s1->ops;
    return exp;
}

static vexexp *build_strexp(char *s)
{
    vexexp *exp = (vexexp*) malloc(sizeof(vexexp));
    assert(exp && "out of memory");
    exp->str = s;
    exp->val = 0;
    exp->ops = NULL;
    return exp;
}

static vexexp *build_numexp(int x)
{
    vexexp *exp = (vexexp*) malloc(sizeof(vexexp));
    assert(exp && "out of memory");
    exp->str = NULL;
    exp->val = x;
    exp->ops = NULL;
    return exp;
}

void yyerror (char *s)
{
   fprintf (stderr,"[FATAL] %s at line %d\n", s, yylineno);
   exit(100);
}

void yywarn (char *s)
{
   fprintf (stderr,"[WARN] %s at line %d\n", s, yylineno);
}

int yywrap ()
{
  return 1;
}
