/* ebner dietmar, 04/2008
   ebner@complang.tuwien.ac.at
*/

#include <iostream>
#include <vector>

/* bgl */
#include <boost/config.hpp>
#include <boost/tuple/tuple.hpp>
#include <boost/graph/adjacency_list.hpp>
#include <boost/graph/graph_utility.hpp>
#include <boost/graph/graphviz.hpp>
#include <boost/graph/reverse_graph.hpp>


using namespace std;
using namespace boost;

#include "sched.h"
extern void print_oplist(oplist *list);


void
sched::ddg_node_label_writer::operator()(std::ostream& out, const ddg_node& v) const {
  std::ostringstream lbl;	
  out << "[";
  if(graph[v].type == START_VERTEX || graph[v].type == END_VERTEX) {
    out << "style=filled fillcolor=red ";
    lbl << (graph[v].type == START_VERTEX ? "S" : "E");
    const ddg_graph_prop &P = get_property(graph, graph_ddg_property);
    if(graph[v].type == START_VERTEX)
      lbl << "\n (LB: " << P.lb << ", UB: " << P.ub << ")";
  }
  else {
    lbl << graph[v].id << ": " << graph[v].instr->as_op << "\n"
	<< "(" << graph[v].earliest << ", " << graph[v].latest << ")";
  }
  out << " label=\"" << lbl.str() << "\"]";
}

void
sched::ddg_edge_label_writer::operator()(std::ostream& out, const ddg_edge& e) const {
  static const char *edge_t_names[MAX_DDG_EDGE_T] =
    {"RAW", "WAW", "WAR", "CTRL" };
  out << "[label=\"" << edge_t_names[graph[e].type] << ", "
      << graph[e].weight << "\"]";
}

void
sched::ddg_graph_property_writer::operator()(std::ostream& out) const {
  out << "graph [bgcolor=lightgrey]" << std::endl;
  out << "node [shape=rectangle style=filled fillcolor=lightblue]" << std::endl;
  out << "edge [style=solid]" << std::endl;
}

sched::ddg_node 
sched::getStartVertex(ddg &DDG) {
  ddg::vertex_iterator I, IE;
  boost::tie(I, IE) = vertices(DDG);
  assert(I != IE && "passed empty graph");
  assert(DDG[*I].type == START_VERTEX && "didn't encounter start vertex where expected");
  return *I;
}
 
sched::ddg_node 
sched::getEndVertex(ddg &DDG) {
  ddg::vertex_iterator I, IE;
  boost::tie(I, IE) = vertices(DDG);
  assert(I != IE && "passed empty graph");
  ++I;
  assert(I != IE && "passed empty graph");
  assert(DDG[*I].type == END_VERTEX && "didn't encounter start vertex where expected");
  return *I;
}

bool
sched::dumpDDG(const char *filename, ddg &DDG) {
  std::ofstream out(filename);
  bool ret = dumpDDG(out, DDG);
  out.close();
  return ret;
}

bool
sched::dumpDDG(std::ostream &out, ddg &DDG) {
  boost::write_graphviz(out, DDG,
			ddg_node_label_writer(DDG),
			ddg_edge_label_writer(DDG),
			ddg_graph_property_writer());
  return true;
}
  
void
sched::computeDDG(string fname, ddg &DDG, vexinstr *istream) {
  vexinstr *l = istream;

  while(l) {
    if(l->is_label)
      cout << "\t" << l->as_op << ":" << endl;
    else {
      cout << "\t" << l->id << ": " << l->as_op << " ";
      print_oplist(l->ops);
      cout << endl;
    }
    l=l->next;
  }

  //TODO
}


void
sched::schedule(ddg &DDG, list<vexinstr*> &) {
  //TODO
}

