#ifndef __SCHED_H_INCLUDED__
#define __SCHED_H_INCLUDED__

#include "vextypes.h"

namespace boost
{
  enum graph_ddg_property_t { graph_ddg_property };
  BOOST_INSTALL_PROPERTY(graph, ddg_property);
}

class sched {
 public:
  enum ddg_edge_t {
    RAW,
    WAW,
    WAR,
    CTRL,
    MAX_DDG_EDGE_T
  };

  enum ddg_op_t {
    MEM,
    REG,
    OTHER,
    MAX_DDG_OP_T
  };
    
  enum ddg_node_t {
    NORMAL,
    START_VERTEX,
    END_VERTEX
  };

  struct ddg_edge_prop {
    unsigned   id;
    int
    weight;
    ddg_edge_t type;
  };
  
  struct ddg_node_prop {
    unsigned id;
    ddg_node_t type;
    vexinstr *instr;
    int earliest;
    int latest;
  };

  //bundled graph properties are not yet supported :-(
  struct ddg_graph_prop {
    unsigned lb;
    unsigned ub;
  };
    
  typedef adjacency_list<listS,           /* OutEdgeList */
    vecS,            /* VertexList */
    bidirectionalS,  /* Directed */
    ddg_node_prop,   /* VertexProperties */
    ddg_edge_prop,   /* EdgeProperties */
    property<graph_ddg_property_t, ddg_graph_prop> /* GraphProperties */
    > ddg;

  typedef graph_traits<ddg>::vertex_descriptor ddg_node;
  typedef graph_traits<ddg>::edge_descriptor ddg_edge;

  typedef std::list< ddg_node > node_list;
  
  /* a functor to dump nodes in graphviz format */
  struct ddg_node_label_writer {
    ddg &graph;
    ddg_node_label_writer(ddg G) : graph(G) { }
    void operator()(std::ostream& out, const ddg_node& v) const;
  };

  /* a functor to dump edges in graphviz format */
  struct ddg_edge_label_writer {
    ddg &graph;
    ddg_edge_label_writer(ddg G) : graph(G) { }
    void operator()(std::ostream& out, const ddg_edge& e) const;
  };

  /* the graph property writer */
  struct ddg_graph_property_writer {
    void operator()(std::ostream& out) const;
  };

  ddg_node getStartVertex(ddg &DDG);
  ddg_node getEndVertex(ddg &DDG);

  bool dumpDDG(std::ostream &, ddg &DDG);
  bool dumpDDG(const char *filename, ddg &DDG);

  void computeDDG(string fname, ddg &DDG, vexinstr* istream);
  void schedule(ddg &DDG, list<vexinstr*> &);
};

#endif
